__author__ = 'Dmitriy Korsakov'

import os

CONFIG_DIRECTORY = os.path.expanduser(os.environ.get("SCALRCLI_HOME", os.path.join(os.path.expanduser("~"), ".scalr")))
CONFIG_PATH = os.path.join(CONFIG_DIRECTORY, "%s.yaml" % os.environ.get("SCALRCLI_PROFILE", "default"))
VERSION = open(os.path.join(os.path.dirname(__file__), "VERSION")).read().strip()
ROUTES_PATH = os.path.join(CONFIG_DIRECTORY, "available_api_routes.json")



__author__ = 'Dmitriy Korsakov'

import os
import json
import pydoc

import yaml

from scalrctl import click
from scalrctl import commands
from scalrctl import settings


class Import(commands.Action):

    action = None
    epilog = "Example: scalr-ctl import < object.yml"

    def _init(self):
        super(Import, self)._init()

        with open(os.path.join(os.path.dirname(__file__), '../scheme/scheme.json')) as fp:
            self.scheme = json.load(fp)

    def get_description(self):
        return "Import Scalr Objects."

    def get_options(self):
        debug = click.Option(('--debug/--no-debug', 'debug'), default=False, help="Print debug messages")
        envid = click.Option(('--envId', 'env_id'), help="Environment ID")
        upd_helpmsg = "Update existing object instead of creating new."
        update = click.Option(('--update', 'update'), is_flag=True, default=False, help=upd_helpmsg, hidden=True)
        dry_run = click.Option(('--dryrun', 'dryrun'), is_flag=True, default=False, help=upd_helpmsg, hidden=True)
        return [debug, update, envid, dry_run]

    def run(self, *args, **kwargs):
        if "debug" in kwargs:
            settings.debug_mode = kwargs.pop("debug")
        if "dryrun" in kwargs:
            dry_run_mode = kwargs.pop("dryrun")

        env_id = kwargs.pop("env_id", None) or settings.envId
        raw = kwargs.pop("raw", None) or click.get_text_stream("stdin")
        obj_data = self._validate_object(raw)

        get_type_route = obj_data["meta"]["scalrctl"]["ROUTE"]
        update_mode = kwargs.pop("update", False)
        http_method = "patch" if update_mode else "post"

        action_scheme = None
        for obj_name, section in self.scheme["export"].items():
            if obj_name != "group_descr" and section["http-method"] == "get" and section["route"] == get_type_route:
                action_scheme = section["%s-params" % http_method]

        if not action_scheme:
            raise click.ClickException("Cannot import Scalr object: API method (%s,%s) not found" % (
                action_scheme["http-method"],
                action_scheme["route"]))

        cls = pydoc.locate(action_scheme['class']) if 'class' in action_scheme else commands.Action
        action = cls(name=self.name, route=action_scheme['route'], http_method=http_method, api_level=self.api_level)

        arguments, kv = obj_data["meta"]["scalrctl"]['ARGUMENTS']

        obj_type = action.get_body_type_params()[0]["name"]
        kv["import-data"] = {obj_type: obj_data["data"]}
        if env_id:
            kv["envId"] = env_id
        if dry_run_mode:
            kv["dryrun"] = dry_run_mode

        click.echo("\x1b[1m%s object %s %s\x1b[0m" % (
            "Updating" if update_mode else "Creating",
            obj_type,
            "ID" if update_mode else ""  # TBD: ID
        ))
        click.echo()

        result = action.run(*arguments, **kv)
        result_json = json.loads(result)

        if "include" in obj_data:
            included_objects = obj_data["include"]

            for obj in included_objects:

                if obj["meta"]["scalrctl"]["ACTION"] == "script-version":  # XXX
                    obj["meta"]["scalrctl"]['ARGUMENTS'][1]['scriptId'] = result_json["data"]["id"]

                inc_raw = yaml.dump(obj)
                inc_kv = {
                    "debug": settings.debug_mode,
                    "env_id": env_id,
                    "dryrun": dry_run_mode,
                    "update": update_mode,
                    "raw": inc_raw,
                }

                self.run(**inc_kv)
        click.echo("%s created." % obj_type)
        click.echo()
        return result

    def _validate_object(self, yml):
        try:
            descr = yaml.safe_load(yml)
            assert "data" in descr
            assert "meta" in descr
            meta = descr["meta"]
            assert "scalrctl" in meta
            info = meta["scalrctl"]
            for key in (
                "METHOD",
                "ROUTE",
                "envId",
                "ARGUMENTS",
                "API_LEVEL"
            ):
                assert key in info

            return descr

        except (Exception, BaseException) as e:

            if settings.debug_mode:
                click.echo(e)
                click.echo(descr)

            raise click.ClickException("Invalid object description")


class ImportImage(commands.Action):

    def pre(self, *args, **kwargs):
        if 'imageId' not in kwargs:
            kwargs["image"] = click.termui.prompt("Image object JSON")

        return super(commands.Action, self).pre(*args, **kwargs)


class UpdateImage(commands.Action):
    mutable_body_parts = ["name"]
    prompt_for = ["imageId"]



__author__ = 'Dmitriy Korsakov'
__doc__ = 'Farm management'
import copy
from scalrctl import commands
from scalrctl import click


class FarmTerminate(commands.Action):

    epilog = "Example: scalr-ctl farm terminate --farmId <ID> --force"

    post_template = {
        "terminateFarmRequest": {"force": True}
    }

    def get_options(self):
        hlp = "It is used to terminate the Server immediately ignoring scalr.system.server_terminate_timeout."
        force_terminate = click.Option(('--force', 'force'), is_flag=True, default=False, help=hlp)
        options = super(FarmTerminate, self).get_options()
        options.append(force_terminate)
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        force = kwargs.pop("force", None)
        post_data = copy.deepcopy(self.post_template)
        post_data["terminateFarmRequest"]["force"] = force
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(FarmTerminate, self).pre(*args, **kv)
        return arguments, kw


class FarmLaunch(commands.Action):

    epilog = "Example: scalr-ctl farm launch --farmId <ID>"
    post_template = {}

    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        kv = {"import-data": {}}
        kv.update(kwargs)
        arguments, kw = super(FarmLaunch, self).pre(*args, **kv)
        return arguments, kw


class FarmClone(commands.Action):

    epilog = "Example: scalr-ctl farm clone --farmId <ID> --name MyNewFarm"
    post_template = {
        "cloneFarmRequest": {"name": ""}
    }

    def get_options(self):
        hlp = "The name of a new Farm."
        hard_terminate = click.Option(('--name', 'name'), required=True, help=hlp)
        options = super(FarmClone, self).get_options()
        options.append(hard_terminate)
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        name = kwargs.pop("name", None)
        post_data = copy.deepcopy(self.post_template)
        post_data["cloneFarmRequest"]["name"] = name
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(FarmClone, self).pre(*args, **kv)
        return arguments, kw



__author__ = 'Dmitriy Korsakov'

import os
import sys
import time
import json
import threading

import yaml
import requests

from scalrctl import click
from scalrctl import defaults
from scalrctl import settings
from scalrctl import commands

SWAGGER_USER_NOUPDATE_TRIGGER = ".noupdate.user"
SWAGGER_USER_FILE = "user.yaml"
SWAGGER_USER_PATH = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_USER_FILE)
SWAGGER_USER_JSONSPEC_FILE = SWAGGER_USER_FILE.split(".")[0] + ".json"
SWAGGER_USER_JSONSPEC_PATH = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_USER_JSONSPEC_FILE)

SWAGGER_ACCOUNT_NOUPDATE_TRIGGER = ".noupdate.account"
SWAGGER_ACCOUNT_FILE = "account.yaml"
SWAGGER_ACCOUNT_PATH = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_ACCOUNT_FILE)
SWAGGER_ACCOUNT_JSONSPEC_FILE = SWAGGER_ACCOUNT_FILE.split(".")[0] + ".json"
SWAGGER_ACCOUNT_JSONSPEC_PATH = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_ACCOUNT_JSONSPEC_FILE)


class UpdateScalrCTL(commands.BaseAction):

    def run(self, *args, **kwargs):
        update()

    def get_description(self):
        return "Fetch new API specification if available."


def is_update_required():
    return not os.path.exists(SWAGGER_USER_PATH) or not os.path.exists(SWAGGER_USER_JSONSPEC_PATH)


def get_spec_url(api_level="user"):
    api_level = api_level or "user"  #XXX: Ugly,  SubCommand class needs to be changed first
    return "{0}://{1}/api/{2}.{3}.yml".format(settings.API_SCHEME, settings.API_HOST, api_level, settings.API_VERSION)


def update():
    """
    Downloads yaml spec and converts it to JSON
    Both files are stored in configuration directory.
    """
    account_paths = []
    user_paths = []

    successfull = False
    text = None
    user_trigger_file = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_USER_NOUPDATE_TRIGGER)

    user_url = get_spec_url(api_level="user")
    user_dst = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_USER_FILE)

    def spinning_cursor():
        while True:
            for cursor in '|/-\\':
                yield cursor

    def draw_spinner(event):
        spinner = spinning_cursor()
        while not event.isSet():
            sys.stdout.write(spinner.next())
            sys.stdout.flush()
            time.sleep(0.1)
            sys.stdout.write('\b')
        sys.stdout.write(' ')
        sys.stdout.flush()

    click.echo("Updating API specifications... ", nl=False)

    e = threading.Event()
    t = threading.Thread(target=draw_spinner, args=(e,))
    t.start()

    try:
        if user_url and not os.path.exists(user_trigger_file):
            # click.echo("Trying to get new UserAPI Spec from %s" % user_url)
            r = requests.get(user_url)

            old = None

            if os.path.exists(user_dst):
                with open(user_dst, "r") as fp:
                    old = fp.read()

            text = r.text

            if text == old:
                # click.echo("UserAPI Spec is already up-to-date.")
                successfull = True
            elif text:
                with open(user_dst, "w") as fp:
                    fp.write(text)
                # click.echo("UserAPI UserSpec successfully updated.")
                successfull = True

        if text or os.path.exists(user_dst):
            struct = yaml.load(text or open(user_dst).read())
            user_paths = struct["paths"].keys()
            json.dump(struct, open(SWAGGER_USER_JSONSPEC_PATH, "w"))


        # Fetch AccountAPI spec and convert to JSON
        text = None
        account_trigger_file = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_ACCOUNT_NOUPDATE_TRIGGER)
        account_url = get_spec_url(api_level="account")
        account_dst = os.path.join(defaults.CONFIG_DIRECTORY, SWAGGER_ACCOUNT_FILE)

        if account_url and not os.path.exists(account_trigger_file):
            # click.echo("Trying to get new AccountAPI Spec from %s" % account_url)
            r = requests.get(account_url)

            old = None

            if os.path.exists(account_dst):
                with open(account_dst, "r") as fp:
                    old = fp.read()

            text = r.text

            if text == old:
                # click.echo("AccountAPI Spec is already up-to-date.")
                successfull = True
            elif text:
                with open(account_dst, "w") as fp:
                    fp.write(text)
                # click.echo("AccountAPI Spec successfully updated.")
                successfull = True

        if text or os.path.exists(account_dst):
            struct = yaml.load(text or open(account_dst).read())
            account_paths = struct["paths"].keys()
            json.dump(struct, open(SWAGGER_ACCOUNT_JSONSPEC_PATH, "w"))

        available_routes = {"user": user_paths, "account": account_paths}
        json.dump(available_routes, open(defaults.ROUTES_PATH, "w"))

    finally:
        e.set()
        t.join()
        if successfull:
            click.echo("Done")
        else:
            click.echo("Failed")


__author__ = 'Dmitriy Korsakov'
__doc__ = 'Server management'
import copy

from scalrctl import commands
from scalrctl import click


class ExecuteScript(commands.Action):

    epilog = "Example: scalr-ctl script execute --serverId <ID> --scriptId <ID> --blocking --timeout 30"

    post_template = {"scriptExecutionRequestObject": {"server": None, "blocking": False, "timeout": None}}

    def get_options(self):
        blocking_hlp = "If it is set Scalr Agent will wait for your Script to finish \
        executing before firing and processing further events."
        blocking = click.Option(('--blocking', 'blocking'), is_flag=True, default=False, help=blocking_hlp)

        server_id_hlp = "Identifier of the Server"
        server_id = click.Option(('--serverId', 'serverId'), required=True, help=server_id_hlp)

        timeout_hlp = "How many secconds should Scalr Agent wait before aborting the execution of this Script."
        timeout = click.Option(('--timeout', 'timeout'), required=False, help=timeout_hlp)

        options = super(ExecuteScript, self).get_options()
        options += [blocking, server_id, timeout]
        return options


    def pre(self, *args, **kwargs):
        """
        before request is made
        """
        blocking = kwargs.pop("blocking", False)
        server_id = kwargs.pop("serverId")
        timeout = kwargs.pop("timeout", False)
        post_data = copy.deepcopy(self.post_template)
        post_data["scriptExecutionRequestObject"]["blocking"] = blocking
        post_data["scriptExecutionRequestObject"]["server"] = server_id
        post_data["scriptExecutionRequestObject"]["timeout"] = timeout
        kv = {"import-data": post_data}
        kv.update(kwargs)
        arguments, kw = super(ExecuteScript, self).pre(*args, **kv)
        return arguments, kw


class ExecuteScriptVersion(ExecuteScript):

    epilog = "Example: scalr-ctl script-version execute \
    --serverId <ID> --scriptId <ID> --scriptVersionNumber <num>--blocking --timeout 30"
